#include <Globals.h>
#include <ConstantPoolTags/Tags.h>
#include <Utils/Utils.h>
#include <cstdint>
#include <cstdlib>
#include <iostream>
#include <ClassFormat.h>
#include <ostream>

void ConstantPoolTags::ConstantPoolParser(ConstantPool &p, JavaClassFormat &jc)
{
	switch (p.tag)
	{
		case Tags::Utf8: {
			// std::cout << "FIXME: Implement Utf8 Constant Pool Tag" << std::endl;
			ConstantUtf8Info ref = static_cast<ConstantUtf8Info>(p.info);
			std::vector<uint8_t> length = Utils::ReadFromStart(Globals::buffer(), 2);
			ref.tag = p.tag;
			ref.length = Utils::vecToVal<uint16_t>(length);

			unsigned char x = Utils::ReadFromStartIntoVal<unsigned char>(Globals::buffer(), 1);
			bool valid = true;

			ref.bytes = new unsigned char[ref.length.ToHostFormat()];

			for (int i = 0; i < ref.length.ToHostFormat(); i++)
			{
				if ((x & 0x80) == 0)
				{
					ref.bytes[i] = x;
				}
				else if ((x & 0x800) == 0)
				{
					unsigned char y = Utils::ReadFromStartIntoVal<unsigned char>(Globals::buffer(), 1);
					ref.bytes[i] = ((x & 0x1f) << 6) + (y & 0x3f);
				}
				else if ((x & 0x10000) == 0)
				{
					unsigned char y = Utils::ReadFromStartIntoVal<unsigned char>(Globals::buffer(), 1);
					unsigned char z = Utils::ReadFromStartIntoVal<unsigned char>(Globals::buffer(), 1);
					ref.bytes[i] = ((x & 0xf) << 12) + ((y & 0x3f) << 6) + (z & 0x3f);
				}
				else
				{
					std::cerr << "Invalid character: " << x << std::endl;
					exit(1);
				}

				if (i != ref.length.ToHostFormat() - 1)
					x = Utils::ReadFromStartIntoVal<unsigned char>(Globals::buffer(), 1);
			}

			p.info = static_cast<ConstantPoolInfo>(ref);
			jc.constant_pool.push_back(p);

			break;
		}
		case Tags::Integer:
			std::cout << "FIXME: Implement Integer Constant Pool Tag" << std::endl;
			exit(EXIT_SUCCESS);
		case Tags::Float:
			std::cout << "FIXME: Implement Float Constant Pool Tag" << std::endl;
			exit(EXIT_SUCCESS);
		case Tags::Long:
			std::cout << "FIXME: Implement Long Constant Pool Tag" << std::endl;
			exit(EXIT_SUCCESS);
		case Tags::Double:
			std::cout << "FIXME: Implement Double Constant Pool Tag" << std::endl;
			exit(EXIT_SUCCESS);
		case Tags::Class: {
			// std::cout << "FIXME: Implement Class Constant Pool Tag" << std::endl;
			ConstantClassInfo ref = static_cast<ConstantClassInfo>(p.info);
			// std::vector<uint8_t> name_index = Utils::ReadFromStart(Globals::buffer(), 2);
			ref.tag = p.tag;
			ref.name_index = Utils::ReadFromStartIntoVal<uint16_t>(Globals::buffer(), 2);

			p.info = static_cast<ConstantPoolInfo>(ref);
			jc.constant_pool.push_back(p);

			// exit(EXIT_SUCCESS);
			break;
		}
		case Tags::String: {
			// std::cout << "FIXME: Implement String Constant Pool Tag" << std::endl;
			ConstantStringInfo ref = static_cast<ConstantStringInfo>(p.info);
			ref.tag = p.tag;
			// std::cout << std::dec << Globals::buffer().size() << std::endl;
			ref.string_index = Utils::ReadFromStartIntoVal<uint16_t>(Globals::buffer(), 2);

			// std::cout << std::dec << ref.string_index.ToHostFormat() << std::endl;
			p.info = static_cast<ConstantPoolInfo>(ref);

			// std::cout << Globals::buffer().size() << std::endl;

			jc.constant_pool.push_back(p);

			// exit(EXIT_SUCCESS);
			break;
		}
		case Tags::Fieldref: {
			// std::cout << "FIXME: Implement Fieldref Constant Pool Tag" << std::endl;
			ConstantFieldrefInfo ref = static_cast<ConstantFieldrefInfo>(p.info);
			ref.tag = p.tag;
			ref.class_index = Utils::ReadFromStartIntoVal<uint16_t>(Globals::buffer(), 2);
			ref.name_and_type_index = Utils::ReadFromStartIntoVal<uint16_t>(Globals::buffer(), 2);

			p.info = static_cast<ConstantPoolInfo>(ref);

			jc.constant_pool.push_back(p);

			// exit(EXIT_SUCCESS);
			break;
		}
		case Tags::Methodref:
		{
			// std::cout << "FIXME: Implement Methodref Constant Pool Tag" << std::endl;
			// ConstantMethodrefInfo ref = *reinterpret_cast<ConstantMethodrefInfo*>(p->info);
			ConstantMethodrefInfo ref = static_cast<ConstantMethodrefInfo>(p.info);

			ref.tag = p.tag;
			ref.class_index = Utils::ReadFromStartIntoVal<uint16_t>(Globals::buffer(), 2);
			ref.name_and_type_index = Utils::ReadFromStartIntoVal<uint16_t>(Globals::buffer(), 2);

			p.info = static_cast<ConstantPoolInfo>(ref);

			jc.constant_pool.push_back(p);

			// exit(EXIT_SUCCESS);
			break;
		}
		case Tags::InterfaceMethodref:
			std::cout << "FIXME: Implement InterfaceMethodref Constant Pool Tag" << std::endl;
			exit(EXIT_SUCCESS);
		case Tags::NameAndType: {
			// std::cout << "FIXME: Implement NameAndType Constant Pool Tag" << std::endl;
			ConstantNameAndTypeInfo ref = static_cast<ConstantNameAndTypeInfo>(p.info);

			ref.tag = p.tag;
			ref.name_index = Utils::ReadFromStartIntoVal<uint16_t>(Globals::buffer(), 2);
			ref.descriptor_index = Utils::ReadFromStartIntoVal<uint16_t>(Globals::buffer(), 2);

			p.info = static_cast<ConstantPoolInfo>(ref);
			jc.constant_pool.push_back(p);

			// exit(EXIT_SUCCESS);
			break;
		}
		case Tags::MethodHandle:
			std::cout << "FIXME: Implement MethodHandle Constant Pool Tag" << std::endl;
			exit(EXIT_SUCCESS);
		case Tags::MethodType:
			std::cout << "FIXME: Implement MethodType Constant Pool Tag" << std::endl;
			exit(EXIT_SUCCESS);
		case Tags::Dynamic:
			std::cout << "FIXME: Implement Dynamic Constant Pool Tag" << std::endl;
			exit(EXIT_SUCCESS);
		case Tags::InvokeDynamic:
			std::cout << "FIXME: Implement InvokeDynamic Constant Pool Tag" << std::endl;
			exit(EXIT_SUCCESS);
		case Tags::Module:
			std::cout << "FIXME: Implement Module Constant Pool Tag" << std::endl;
			exit(EXIT_SUCCESS);
		case Tags::Package:
			std::cout << "FIXME: Implement Package Constant Pool Tag" << std::endl;
			exit(EXIT_SUCCESS);
		default:
			std::cerr << "Unknown tag: " << std::hex << (int)p.tag << " @ byte " << std::dec << Globals::buffer().size() << " from end of class file" << std::endl;
			exit(EXIT_FAILURE);
	}
}